use crate::util::index::Ix2;

fn solve(
    input: &str,
    mut start: Ix2,
    within_fn: fn(&Ix2) -> bool,
    lookup_fn: fn(&Ix2) -> char,
) -> String {
    input
        .lines()
        .fold(Vec::new(), |mut acc, line| {
            for dir in line.bytes() {
                let newpos = start + dir.into();
                if (within_fn)(&newpos) {
                    start = newpos;
                }
            }
            acc.push((lookup_fn)(&start).to_string());
            acc
        })
        .join("")
}

pub fn part1(input: &str) -> impl std::fmt::Display {
    #[inline]
    #[rustfmt::skip]
    fn within(coord: &Ix2) -> bool {
        [
            [0, 0], [1, 0], [2, 0],
            [0, 1], [1, 1], [2, 1],
            [0, 2], [1, 2], [2, 2]
        ].contains(&coord.0)
    }

    #[inline]
    fn lookup(coord: &Ix2) -> char {
        match coord.0 {
            [0, 0] => '7',
            [1, 0] => '8',
            [2, 0] => '9',
            [0, 1] => '4',
            [1, 1] => '5',
            [2, 1] => '6',
            [0, 2] => '1',
            [1, 2] => '2',
            [2, 2] => '3',
            _ => unreachable!(""),
        }
    }
    solve(input, [0, 2].into(), within, lookup)
}

pub fn part2(input: &str) -> impl std::fmt::Display {
    #[inline]
    #[rustfmt::skip]
    fn within(coord: &Ix2) -> bool {
        [
                            [2, 0],
                    [1, 1], [2, 1], [3, 1],
            [0, 2], [1, 2], [2, 2], [3, 2], [4, 2],
                    [1, 3], [2, 3], [3, 3],
                            [2, 4],
        ].contains(&coord.0)
    }

    #[inline]
    fn lookup(coord: &Ix2) -> char {
        match coord.0 {
            [2, 0] => 'D',
            [1, 1] => 'A',
            [2, 1] => 'B',
            [3, 1] => 'C',
            [0, 2] => '5',
            [1, 2] => '6',
            [2, 2] => '7',
            [3, 2] => '8',
            [4, 2] => '9',
            [1, 3] => '2',
            [2, 3] => '3',
            [3, 3] => '4',
            [2, 4] => '1',
            _ => unreachable!(""),
        }
    }
    solve(input, [0, 2].into(), within, lookup)
}

#[test]
fn test_part1() {
    assert_eq!("1985", part1("ULL\nRRDDD\nLURDL\nUUUUD").to_string())
}

#[test]
fn test_part2() {
    assert_eq!("5DB3", part2("ULL\nRRDDD\nLURDL\nUUUUD").to_string())
}
